// ************************************************************************************************
//
// PuzzleClock Nixie clock using the PIC 16F627A.
//
// Features:
//
// (x) displays time via 6 Nixie tubes
// (x) time is adjustable via two pushbuttons (hours and minutes, respectively)
// (x) overall brightness pulses between 0-100% @ 1Hz
// (x) the positions of hours, minutes and seconds (both ones and tens) change each second,
//     (adjust this using the arrays in lines 41-46.
//
// http://www.jb-electronics.de/html/elektronik/nixies/n_puzzleclock.htm
//
// copyright www.jb-electronics.de
//
// ************************************************************************************************

// include PIC header with predefined ports and registers
#include <pic.h>

// ************************************************************************************************
// method prototypes
void updatePWM ();
void updateDisplay ();

// ************************************************************************************************
// global variables
static unsigned char seconds, minutes, hours, second_counter, MIN_TIMEOUT, HRS_TIMEOUT;
static int DUTY_CYCLE, timebase, SW_MIN_BUFFER, SW_HRS_BUFFER;

// connection from LS74141 to Nixie tubes: use the getBCD array to keep the connections from the LS74141 to the Nixie tube flexible.

// for 1:1 wiring as depicted on the website, use this arrray:
static unsigned char getBCD[10] = {0b0000, 0b0001, 0b0010, 0b0011, 0b0100, 0b0101, 0b0110, 0b0111, 0b1000, 0b1001};

// just as a note: in my setup, I used the following array:
// static unsigned char getBCD[10] = {0b0001, 0b0000, 0b1010, 0b0010, 0b1000, 0b0110, 0b1110, 0b1100, 0b0100, 0b1001};

// these arrays determine the position of the hours, minutes and seconds within the six available Nixie tubes.
static bank1 unsigned char getPosSec1[10]  = {5, 4, 1, 2, 4, 1, 0, 1, 3, 3};
static bank1 unsigned char getPosSec10[10] = {4, 3, 2, 5, 2, 2, 3, 4, 4, 4};
static bank1 unsigned char getPosMin1[10]  = {2, 1, 3, 4, 3, 5, 5, 0, 0, 1};
static bank1 unsigned char getPosMin10[10] = {3, 5, 4, 3, 5, 3, 2, 2, 5, 2};
static bank1 unsigned char getPosHrs1[10]  = {0, 2, 0, 1, 0, 0, 4, 3, 1, 5};
static bank1 unsigned char getPosHrs10[10] = {1, 0, 5, 0, 1, 4, 1, 5, 2, 0};


// ************************************************************************************************
// main method
void main (void) {


    // ********************************************************************************************
	// configuration word (XT oscillator, no WDT, Power-Up-Timer yes, MCLR yes, Brown-Out-Detect yes, LVP no, no code and data protection)
	__CONFIG(0x3F61);


	// ********************************************************************************************
	// configure IO settings

	// DATA for the shift registers
	#define DATA	RB0
	TRISB0 = 0;

	// CLK for the shift registers
	#define CLK		RB1
	TRISB1 = 0;

	// STROBE for the shift registers
	#define STROBE	RB2
	TRISB2 = 0;

	// will be used for PWMing the Nixie tubes
	#define ENABLE	RB3
	TRISB3 = 0;

	// set hours
	#define SW_HRS	RA2
	TRISA2 = 1;

	// set minutes
	#define SW_MIN	RA3
	TRISA3 = 1;

	// disable comparator module such that RA3 can be used as an ordinary input
	CMCON = 0b111;


	// ********************************************************************************************
	// configure Timer0 (internal clock, no prescaler)
	T0CS = 0;
	PSA = 1;

	// interrupt on timer overflow
	T0IE = 1;

	// enable global interrupts
	GIE = 1;


	// ********************************************************************************************
	// configure the PWM module

	// upper limit for Timer2 is 255 (8 bit):
	PR2 = 0xff;

	// set duty cycle value (between 0...1023)
	// the registers are set in the method updatePWM(), the variable DUTY_CYCLE is not a register, but merely a variable used in this program
	DUTY_CYCLE = 0;

	// set Timer2 prescaler to 1:1 (00 = 1:1, 01 = 1:4, 1x = 1:16)
	// and start Timer2 (bit no. 2, MSB)
	T2CON = 0b100;

	// Set CCP module to PWM mode
	CCP1CON = 0b1100;

	
	// ********************************************************************************************
	// other stuff (i.e. preset variables)

	hours 	= 12;
	minutes = 0;
	seconds = 0;
	second_counter = 0;


	// ********************************************************************************************
	// main loop
	while (1) {


		// ****************************************************************************************
		// send the time
		updateDisplay ();


		// ****************************************************************************************
		// set the time?
		
		// hours button pressed?
		if ((SW_HRS == 1) && (SW_HRS_BUFFER == 0) && (MIN_TIMEOUT == 0)) {

			// debounce the hours switch
			SW_HRS_BUFFER = 1000;
	
			// advance the minutes
			hours++;
			if (hours >= 24) {
				hours = 0;
			}

			// set the hours menu timeout to 3 seconds; exit the minutes menu
			HRS_TIMEOUT = 3;
			MIN_TIMEOUT = 0;

		}

		// minutes button pressed?
		if ((SW_MIN == 1) && (SW_MIN_BUFFER == 0) && (HRS_TIMEOUT == 0)) {

			// debounce the minute switch
			SW_MIN_BUFFER = 1000;
	
			// advance the minutes
			minutes++;
			if (minutes >= 60) {
				minutes = 0;
			}

			// set the minute menu timeout to 3 seconds; exit the hours menu
			MIN_TIMEOUT = 3;
			HRS_TIMEOUT = 0;

		}


	}


}


// ************************************************************************************************
// ISR generates accurate 1Hz clock (if a 4.194304MHz crystal is used)
static void interrupt isr  (void) {


	// Timer0 interrupt
	if (T0IF) {


		// lets the output brightness pulse periodically
		// note: when the timebase overflows (i.e. a new second begins), the display is dark such that the digit flip will not be visible
		DUTY_CYCLE = timebase >> 2;
		if (timebase > 2048) {
			DUTY_CYCLE = 2048 - DUTY_CYCLE;
		}
		updatePWM ();


		// create 1Hz (called 4096 times per second when using a 4.194304MHz crystal)
		timebase++;
		if (timebase >= 4095) {

			seconds++;
			if (seconds >= 60) {
				seconds = 0;
				minutes++;
				if (minutes >= 60) {
					minutes = 0;
					hours++;
					if (hours >= 24) {
						hours = 0;
					}
				}
			}

			// countdown of menu timeouts
			if (MIN_TIMEOUT > 0) {
				MIN_TIMEOUT--;
			}
			if (HRS_TIMEOUT > 0) {
				HRS_TIMEOUT--;
			}
			timebase = 0;
		}

		// clear buffers
		if (SW_MIN_BUFFER > 0) {
			SW_MIN_BUFFER--;
		}
		if (SW_HRS_BUFFER > 0) {
			SW_HRS_BUFFER--;
		}

		// reset interrupt flag so that this method won't be called just again
		T0IF = 0;

	}

}


// ************************************************************************************************
// this method writes the new DC value into the respective registers.
void updatePWM () {

	// hi-part
	CCPR1L  = DUTY_CYCLE >> 2;
	
	// two LSBs in the config register
	CCP1CON = 0b1100 | ((DUTY_CYCLE & 0b11) << 2);

}


// ************************************************************************************************
// sends the data to the shift registers
void updateDisplay () {


	// variables
	char i, j, tmp1, tmp10;
	char values[6] = {0b1111, 0b1111, 0b1111, 0b1111, 0b1111, 0b1111};


	// we only have ten separate position configurations here (instead of the 6! = 720 thinkable)
	// therefore it is OK to identify the second_counter with the ones of the seconds
	// (otherwise this counter had to be independently programmed in the ISR)
	second_counter = seconds % 10;


	// insert time data into the arrays


	// hours
	tmp1 = getPosHrs1[second_counter];
	tmp10 = getPosHrs10[second_counter];
	if (MIN_TIMEOUT == 0) {
		values[tmp1]  = getBCD[hours % 10];
		values[tmp10] = getBCD[hours / 10];
	}


	// minutes
	tmp1 = getPosMin1[second_counter];
	tmp10 = getPosMin10[second_counter];
	if (HRS_TIMEOUT == 0) {
		values[tmp1]  = getBCD[minutes % 10];
		values[tmp10] = getBCD[minutes / 10];
	}


	// seconds
	tmp1 = getPosSec1[second_counter];
	tmp10 = getPosSec10[second_counter];
	if ((HRS_TIMEOUT == 0) && (MIN_TIMEOUT == 0)) {
		values[tmp1]  = getBCD[second_counter];
		values[tmp10] = getBCD[seconds / 10];
	}


	// initialize
	STROBE = 0;


	// send the data in each Nixie tube (index i)
	for (i = 0; i <= 5; i++) {

		// send each bit (index j)
		for (j = 0; j <= 3; j++) {
			DATA = (values[i] & (1 << j)) >> j;
			CLK = 1;
			CLK = 0;
		}

	}


	// STROBE the registers
	STROBE = 1;
	STROBE = 0;


}
